// the file was modified to use program memory to store data
// also CGI functionality was enhanced to allow parsing url parameters
//  Vladimir S. Fonov ~ vladimir.fonov <at> gmail.com

/*
 * Copyright (c) 2004, Adam Dunkels.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file is part of the uIP TCP/IP stack.
 *
 * Author: Adam Dunkels <adam@sics.se>
 *
 * $Id: httpd.c,v 1.2 2006/06/11 21:46:38 adam Exp $
 */ 


#include "uip.h"
#include "psock.h"
#include "httpd.h"
#include "httpd-cgi.h"
#include "httpd-fs.h"
#include "ee_settings.h"

#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h> 
#include <string.h>

HTTPD_CGI_CALL(cgi_index,    "index",    run_index);
HTTPD_CGI_CALL(cgi_action,   "action",   run_action);

static const struct httpd_cgi_call *calls[] = 
{ 
	&cgi_index,
	&cgi_action,
	NULL 
};

uint8_t http_get_parameters_parse(char *par,uint8_t mx)
{
	uint8_t count=0;
	uint8_t i=0;
	for(;par[i]&&i<mx;i++)
	{
		if(par[i]=='=') 
		{
			count++;
			par[i]=0;
		} else if(par[i]=='&')
		  	par[i]=0;
	}
	return count;
}

char * http_get_parameter_name(char *par,uint8_t cnt,uint8_t mx)
{
	uint8_t i,j;
	cnt*=2;
	for(i=0,j=0;j<mx&&i<cnt;j++) {
		if(!par[j]) 
			i++;
	}

	return j==mx?"":par+j;
}

char * http_get_parameter_value(char *par,uint8_t cnt,uint8_t mx)
{
	uint8_t i,j;
	cnt*=2;
	cnt++;
	for(i=0,j=0;j<mx&&i<cnt;j++) {
		if(!par[j])
			i++;
	}

	return j==mx?"":par+j;
}

void http_url_decode(const char *in,char *out,uint8_t mx)
{
	uint8_t i,j;
	char tmp[3]={0,0,0};
	for(i=0,j=0;j<mx&&in[i];i++)
	{
		if(in[i]=='%') {
			tmp[0]=in[++i];
			tmp[1]=in[++i];
			out[j++]=(char)strtol(tmp,NULL,16);
		} else 
			out[j++]=in[i];
	}
}


static PT_THREAD(nullfunction(struct httpd_state *s, PGM_P ptr))
{
	PSOCK_BEGIN(&s->sout);
	PSOCK_SEND_PSTR(&s->sout,PSTR("Not Implemented"));
	PSOCK_END(&s->sout);
}

httpd_cgifunction httpd_cgi(char *name)
{
	const struct httpd_cgi_call **f;

	#if DEBUG_SERIAL
		printf_P(PSTR("Calling CGI script:%s\r\n"),name);
	#endif

	/* Find the matching name in the table, return the function. */
	for(f = calls; *f != NULL; ++f)
	{
		if(strncmp((*f)->name, name, strlen((*f)->name)) == 0) {
			return (*f)->function;
		}
	}

	#if DEBUG_SERIAL
		printf_P(PSTR("Script not found\r\n"));
	#endif

	return nullfunction;
}

httpd_cgifunction httpd_cgi_P(PGM_P name)
{
	const struct httpd_cgi_call **f;

	#if DEBUG_SERIAL
		printf_P(PSTR("Calling CGI script from program memory\r\n"));
	#endif

	/* Find the matching name in the table, return the function. */
	for(f = calls; *f != NULL; ++f)
	{
		if(strncmp_P((*f)->name, name, strlen((*f)->name)) == 0) {
			return (*f)->function;
		}
	}
	#if DEBUG_SERIAL
		printf_P(PSTR("Script not found\r\n"));
	#endif
	return nullfunction;
}


static PT_THREAD(run_action(struct httpd_state *s, PGM_P ptr))
{
	static uint8_t pcount;

	PSOCK_BEGIN(&s->sout);

	//check if there are parameters passed 
	if(s->param[0] && (pcount=http_get_parameters_parse(s->param,sizeof(s->param)))>0)
	{
		static uint8_t i;
		//walk through parameters
		for(i=0;i<pcount;i++)
		{
			static char *pname,*pval;
			pname=http_get_parameter_name(s->param,i,sizeof(s->param));
			pval =http_get_parameter_value(s->param,i,sizeof(s->param));
			// Found action parameter
			if(!strcmp_P(pname,PSTR("action")))
			{
				// Action to turn power on
				if(!strcmp_P(pval,PSTR("on")))
					PORTA = (1<<PORTA7);	
				// Action to turn power off
				if(!strcmp_P(pval,PSTR("off")))
					PORTA = (0<<PORTA7);
				break;
			}
		}
	}

	PSOCK_SEND_PSTR(&s->sout,PSTR("<meta http-equiv=\"refresh\" content=\"0; url=index.shtml\">"));

	PSOCK_END(&s->sout);
}

 
static PT_THREAD(run_index(struct httpd_state *s, PGM_P ptr))
{
	static uint8_t power_on;

	power_on = ((PINA & (1<<PORTA7)) >> PORTA7);


	PSOCK_BEGIN(&s->sout);

	PSOCK_SEND_PSTR(&s->sout,PSTR("<h1>Power Manager</h1><br>"));
	
	if (power_on) {
		PSOCK_SEND_PSTR(&s->sout,PSTR("\
		<h3><font color=\"green\">Power is on</font></h3>\
		<form method=\"get\" action=\"action.shtml\">\
		<input type=\"hidden\" name=\"action\" value=\"off\" />\
		<input type=\"submit\" value=\"Turn Off\" />\
		</form>\
		"));	
	}
	else {
		PSOCK_SEND_PSTR(&s->sout,PSTR("\
		<h3><font color=\"red\">Power is off</font></h3>\
		<form method=\"get\" action=\"action.shtml\">\
		<input type=\"hidden\" name=\"action\" value=\"on\" />\
		<input type=\"submit\" value=\"Turn On\" />\
		</form>\
		"));
	}

	PSOCK_END(&s->sout);
}
